def private_function(name):
    """private_function is a docstring().

    :meta private:
    """

def _public_function(name):
    """public_function is a docstring().

    :meta public:
    """


PRIVATE_CONSTANT = None  #: :meta private:
_PUBLIC_CONSTANT = None  #: :meta public:
